/*
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "src/include/processor.h"
#include "src/include/types.h"

#include <chrono>
#include <iomanip>
#include <iostream>
#include <map>

using namespace ppds;
constexpr uint16_t ITERATIONS = 5;

PlanNode Q1_1LQP(const std::string& param_1 = "1", const std::string& param_2 = "3") {
    PlanNode get_lo(NodeType::TableScan, {{"file_path", "data/benchmark/lineorder.tbl"}}, {});
    PlanNode proj_lo(NodeType::Projection, {{"column_ids", "5,8,11,13"}}, {get_lo});
    PlanNode pred_lo_1(NodeType::Predicate, {{"column_id", "2"}, {"condition", ">="}, {"literal", param_1}}, {proj_lo});
    PlanNode pred_lo_2(NodeType::Predicate, {{"column_id", "2"}, {"condition", "<="}, {"literal", param_2}}, {pred_lo_1});
    PlanNode pred_lo_3(NodeType::Predicate, {{"column_id", "1"}, {"condition", ">"}, {"literal", "25"}}, {pred_lo_2});

    PlanNode get_d(NodeType::TableScan, {{"file_path", "data/benchmark/date.tbl"}}, {});
    PlanNode proj_d(NodeType::Projection, {{"column_ids", "0,4"}}, {get_d});
    PlanNode pred_d_1(NodeType::Predicate, {{"column_id", "1"}, {"condition", "="}, {"literal", "1993"}}, {proj_d});

    PlanNode join_1(NodeType::EquiJoin, {{"left_column_id", "0"}, {"right_column_id", "0"}}, {pred_lo_3, pred_d_1});
    PlanNode gb_1(NodeType::GroupBy, {{"aggregate_column_id", "3"}, {"aggregate_type", "avg"}}, {join_1});

    return gb_1;
}

PlanNode Q1_2LQP(const std::string& param_1 = "26", const std::string& param_2 = "35") {
    PlanNode get_lo(NodeType::TableScan, {{"file_path", "data/benchmark/lineorder.tbl"}}, {});
    PlanNode proj_lo(NodeType::Projection, {{"column_ids", "5,8,11,13"}}, {get_lo});
    PlanNode pred_lo_1(NodeType::Predicate, {{"column_id", "2"}, {"condition", ">="}, {"literal", "4"}}, {proj_lo});
    PlanNode pred_lo_2(NodeType::Predicate, {{"column_id", "2"}, {"condition", "<="}, {"literal", "6"}}, {pred_lo_1});
    PlanNode pred_lo_3(NodeType::Predicate, {{"column_id", "1"}, {"condition", ">="}, {"literal", param_1}}, {pred_lo_2});
    PlanNode pred_lo_4(NodeType::Predicate, {{"column_id", "1"}, {"condition", "<="}, {"literal", param_2}}, {pred_lo_3});

    PlanNode get_d(NodeType::TableScan, {{"file_path", "data/benchmark/date.tbl"}}, {});
    PlanNode proj_d(NodeType::Projection, {{"column_ids", "0,5"}}, {get_d});
    PlanNode pred_d_1(NodeType::Predicate, {{"column_id", "1"}, {"condition", "="}, {"literal", "199401"}}, {proj_d});

    PlanNode join_1(NodeType::EquiJoin, {{"left_column_id", "0"}, {"right_column_id", "0"}}, {pred_lo_4, pred_d_1});
    PlanNode gb_1(NodeType::GroupBy, {{"aggregate_column_id", "3"}, {"aggregate_type", "avg"}}, {join_1});

    return gb_1;
}

PlanNode Q2_1LQP(const std::string& param_1 = "MFGR#12") {
    PlanNode get_lo(NodeType::TableScan, {{"file_path", "data/benchmark/lineorder.tbl"}}, {});
    PlanNode proj_lo(NodeType::Projection, {{"column_ids", "3,4,5,12"}}, {get_lo});

    PlanNode get_d(NodeType::TableScan, {{"file_path", "data/benchmark/date.tbl"}}, {});
    PlanNode proj_d(NodeType::Projection, {{"column_ids", "0,4"}}, {get_d});

    PlanNode get_p(NodeType::TableScan, {{"file_path", "data/benchmark/part.tbl"}}, {});
    PlanNode proj_p(NodeType::Projection, {{"column_ids", "0,3"}}, {get_p});
    PlanNode pred_p_1(NodeType::Predicate, {{"column_id", "1"}, {"condition", "="}, {"literal", param_1}}, {proj_p});

    PlanNode get_s(NodeType::TableScan, {{"file_path", "data/benchmark/supplier.tbl"}}, {});
    PlanNode proj_s(NodeType::Projection, {{"column_ids", "0,5"}}, {get_s});
    PlanNode pred_s_1(NodeType::Predicate, {{"column_id", "1"}, {"condition", "="}, {"literal", "AMERICA"}}, {proj_s});

    PlanNode join_1(NodeType::EquiJoin, {{"left_column_id", "2"}, {"right_column_id", "0"}}, {proj_lo, proj_d,});
    PlanNode join_2(NodeType::EquiJoin, {{"left_column_id", "0"}, {"right_column_id", "0"}}, {join_1, pred_p_1});
    PlanNode join_3(NodeType::EquiJoin, {{"left_column_id", "1"}, {"right_column_id", "0"}}, {join_2, pred_s_1});

    PlanNode gb_1(NodeType::GroupBy, {{"group_by_column_id", "5"}, {"aggregate_column_id", "3"}, {"aggregate_type", "max"}}, {join_3});

    return gb_1;
}

PlanNode Q2_2LQP(const std::string& param_1 = "MFGR#2239") {
    PlanNode get_lo(NodeType::TableScan, {{"file_path", "data/benchmark/lineorder.tbl"}}, {});
    PlanNode proj_lo(NodeType::Projection, {{"column_ids", "3,4,5,12"}}, {get_lo});

    PlanNode get_d(NodeType::TableScan, {{"file_path", "data/benchmark/date.tbl"}}, {});
    PlanNode proj_d(NodeType::Projection, {{"column_ids", "0,4"}}, {get_d});

    PlanNode get_p(NodeType::TableScan, {{"file_path", "data/benchmark/part.tbl"}}, {});
    PlanNode proj_p(NodeType::Projection, {{"column_ids", "0,4"}}, {get_p});
    PlanNode pred_p_1(NodeType::Predicate, {{"column_id", "1"}, {"condition", "="}, {"literal", param_1}}, {proj_p});

    PlanNode get_s(NodeType::TableScan, {{"file_path", "data/benchmark/supplier.tbl"}}, {});
    PlanNode proj_s(NodeType::Projection, {{"column_ids", "0,5"}}, {get_s});
    PlanNode pred_s_1(NodeType::Predicate, {{"column_id", "1"}, {"condition", "="}, {"literal", "EUROPE"}}, {proj_s});

    PlanNode join_1(NodeType::EquiJoin, {{"left_column_id", "2"}, {"right_column_id", "0"}}, {proj_lo, proj_d,});
    PlanNode join_2(NodeType::EquiJoin, {{"left_column_id", "0"}, {"right_column_id", "0"}}, {join_1, pred_p_1});
    PlanNode join_3(NodeType::EquiJoin, {{"left_column_id", "1"}, {"right_column_id", "0"}}, {join_2, pred_s_1});

    PlanNode gb_1(NodeType::GroupBy, {{"group_by_column_id", "5"}, {"aggregate_column_id", "3"}, {"aggregate_type", "min"}}, {join_3});

    return gb_1;
}

int main() {
    std::cout << "###    \033[1mBENCHMARK\033[0m    ###" << std::endl;
    std::cout << "-----------------------" << std::endl;
    std::cout << "QueryID | Run |  Time" << std::endl;
    std::cout << "-----------------------" << std::endl;
    std::map<std::string, PlanNode> lqps{{"Q1.1", Q1_1LQP()},
                                         {"Q1.2", Q1_2LQP()},
                                         {"Q2.1", Q2_1LQP()},
                                         {"Q2.2", Q2_2LQP()}};
    double total_time = 0;
    for (auto& lqp : lqps) {
        for (uint16_t i = 0; i < ITERATIONS; i++) {
            const auto begin = std::chrono::system_clock::now();
            auto result = Processor::ExecuteQuery(lqp.second);
            const auto end = std::chrono::system_clock::now();
            std::chrono::duration<double> duration = end - begin;
            total_time += duration.count();

            std::cout << lqp.first << "    |  " << std::setprecision(3) << i << "  | " << duration.count() << " s" << std::endl;
        }
        std::cout << "-----------------------" << std::endl;
    }

    std::cout << "\n\033[1mTOTAL TIME: " << std::setprecision(5) << total_time << " s\033[0m" << std::endl;
}
